@@ -11,18 +11,22 @@ use super::{
     initialize_instance, initialize_vmcontext, InstanceAllocationRequest, InstanceAllocator,
     InstanceHandle, InstantiationError,
 };
-use crate::{instance::Instance, Memory, Mmap, Table, VMContext};
+use crate::MemFdSlot;
+use crate::{instance::Instance, Memory, Mmap, ModuleMemFds, Table};
 use anyhow::{anyhow, bail, Context, Result};
-use rand::Rng;
+use libc::c_void;
 use std::convert::TryFrom;
-use std::marker;
 use std::mem;
-use std::sync::{Arc, Mutex};
+use std::sync::Arc;
+use std::sync::Mutex;
 use wasmtime_environ::{
-    EntitySet, HostPtr, MemoryStyle, Module, PrimaryMap, Tunables, VMOffsets, VMOffsetsFields,
+    HostPtr, MemoryIndex, MemoryStyle, Module, PrimaryMap, Tunables, VMOffsets, VMOffsetsFields,
     WASM_PAGE_SIZE,
 };
 
+mod index_allocator;
+use index_allocator::{PoolingAllocationState, SlotId};
+
 cfg_if::cfg_if! {
     if #[cfg(windows)] {
         mod windows;
@@ -248,20 +252,19 @@ pub enum PoolingAllocationStrategy {
     NextAvailable,
     /// Allocate from a random available instance.
     Random,
-}
-
-impl PoolingAllocationStrategy {
-    fn next(&self, free_count: usize) -> usize {
-        debug_assert!(free_count > 0);
-
-        match self {
-            Self::NextAvailable => free_count - 1,
-            Self::Random => rand::thread_rng().gen_range(0..free_count),
-        }
-    }
+    /// Try to allocate an instance slot that was previously used for
+    /// the same module, potentially enabling faster instantiation by
+    /// reusing e.g. memory mappings.
+    ReuseAffinity,
 }
 
 impl Default for PoolingAllocationStrategy {
+    #[cfg(feature = "memfd-allocator")]
+    fn default() -> Self {
+        Self::ReuseAffinity
+    }
+
+    #[cfg(not(feature = "memfd-allocator"))]
     fn default() -> Self {
         Self::NextAvailable
     }
@@ -281,14 +284,14 @@ struct InstancePool {
     mapping: Mmap,
     instance_size: usize,
     max_instances: usize,
-    free_list: Mutex<Vec<usize>>,
+    index_allocator: Mutex<PoolingAllocationState>,
     memories: MemoryPool,
     tables: TablePool,
-    empty_module: Arc<Module>,
 }
 
 impl InstancePool {
     fn new(
+        strategy: PoolingAllocationStrategy,
         module_limits: &ModuleLimits,
         instance_limits: &InstanceLimits,
         tunables: &Tunables,
@@ -329,17 +332,11 @@ impl InstancePool {
             mapping,
             instance_size,
             max_instances,
-            free_list: Mutex::new((0..max_instances).collect()),
+            index_allocator: Mutex::new(PoolingAllocationState::new(strategy, max_instances)),
             memories: MemoryPool::new(module_limits, instance_limits, tunables)?,
             tables: TablePool::new(module_limits, instance_limits)?,
-            empty_module: Arc::new(Module::default()),
         };
 
-        // Use a default module to initialize the instances to start
-        for i in 0..instance_limits.count as usize {
-            pool.initialize(module_limits, i);
-        }
-
         Ok(pool)
     }
 
@@ -348,42 +345,28 @@ impl InstancePool {
         &mut *(self.mapping.as_mut_ptr().add(index * self.instance_size) as *mut Instance)
     }
 
-    fn initialize(&self, limits: &ModuleLimits, index: usize) {
-        unsafe {
-            let instance = self.instance(index);
-
-            // Write a default instance with preallocated memory/table map storage to the ptr
-            std::ptr::write(
-                instance as _,
-                Instance {
-                    module: self.empty_module.clone(),
-                    offsets: VMOffsets::new(HostPtr, &self.empty_module),
-                    memories: PrimaryMap::with_capacity(limits.memories as usize),
-                    tables: PrimaryMap::with_capacity(limits.tables as usize),
-                    dropped_elements: EntitySet::new(),
-                    dropped_data: EntitySet::new(),
-                    host_state: Box::new(()),
-                    wasm_data: &[],
-                    vmctx_initialized: false,
-                    vmctx: VMContext {
-                        _marker: marker::PhantomPinned,
-                    },
-                },
-            );
-        }
-    }
-
     unsafe fn setup_instance(
         &self,
         index: usize,
         mut req: InstanceAllocationRequest,
     ) -> Result<InstanceHandle, InstantiationError> {
+        let host_state = std::mem::replace(&mut req.host_state, Box::new(()));
+        let instance_data = Instance::create_raw(
+            &req.module,
+            req.unique_id,
+            &*req.wasm_data,
+            PrimaryMap::default(),
+            PrimaryMap::default(),
+            host_state,
+        );
+
+        // Instances are uninitialized memory at first; we need to
+        // write an empty but initialized `Instance` struct into the
+        // chosen slot before we do anything else with it. (This is
+        // paired with a `drop_in_place` in deallocate below.)
         let instance = self.instance(index);
 
-        instance.module = req.module.clone();
-        instance.offsets = VMOffsets::new(HostPtr, instance.module.as_ref());
-        instance.host_state = std::mem::replace(&mut req.host_state, Box::new(()));
-        instance.wasm_data = &*req.wasm_data;
+        std::ptr::write(instance as _, instance_data);
 
         // set_instance_memories and _tables will need the store before we can completely
         // initialize the vmcontext.
@@ -392,8 +375,10 @@ impl InstancePool {
         }
 
         Self::set_instance_memories(
+            index,
             instance,
-            self.memories.get(index),
+            &self.memories,
+            req.memfds,
             self.memories.max_wasm_pages,
         )?;
 
@@ -412,16 +397,14 @@ impl InstancePool {
 
     fn allocate(
         &self,
-        strategy: PoolingAllocationStrategy,
         req: InstanceAllocationRequest,
     ) -> Result<InstanceHandle, InstantiationError> {
         let index = {
-            let mut free_list = self.free_list.lock().unwrap();
-            if free_list.is_empty() {
+            let mut alloc = self.index_allocator.lock().unwrap();
+            if alloc.is_empty() {
                 return Err(InstantiationError::Limit(self.max_instances as u32));
             }
-            let free_index = strategy.next(free_list.len());
-            free_list.swap_remove(free_index)
+            alloc.alloc(req.unique_id).index()
         };
 
         unsafe {
@@ -449,20 +432,44 @@ impl InstancePool {
         let instance = unsafe { &mut *handle.instance };
 
         // Decommit any linear memories that were used
-        for (memory, base) in instance.memories.values_mut().zip(self.memories.get(index)) {
+        for ((def_mem_idx, memory), base) in
+            instance.memories.iter_mut().zip(self.memories.get(index))
+        {
             let mut memory = mem::take(memory);
             debug_assert!(memory.is_static());
 
-            // Reset any faulted guard pages as the physical memory may be reused for another instance in the future
-            #[cfg(all(feature = "uffd", target_os = "linux"))]
-            memory
-                .reset_guard_pages()
-                .expect("failed to reset guard pages");
-            drop(&mut memory); // require mutable on all platforms, not just uffd
+            match memory {
+                Memory::Static {
+                    memfd_slot: Some(mut memfd_slot),
+                    ..
+                } => {
+                    let mem_idx = instance.module.memory_index(def_mem_idx);
+                    // If there was any error clearing the memfd, just
+                    // drop it here, and let the drop handler for the
+                    // MemFdSlot unmap in a way that retains the
+                    // address space reservation.
+                    if memfd_slot.clear_and_remain_ready().is_ok() {
+                        self.memories.return_memfd_slot(index, mem_idx, memfd_slot);
+                    }
+                }
 
-            let size = memory.byte_size();
-            drop(memory);
-            decommit_memory_pages(base, size).expect("failed to decommit linear memory pages");
+                _ => {
+                    // Reset any faulted guard pages as the physical
+                    // memory may be reused for another instance in
+                    // the future.
+                    #[cfg(all(feature = "uffd", target_os = "linux"))]
+                    memory
+                        .reset_guard_pages()
+                        .expect("failed to reset guard pages");
+                    // require mutable on all platforms, not just uffd
+                    drop(&mut memory);
+
+                    let size = memory.byte_size();
+                    drop(memory);
+                    decommit_memory_pages(base, size)
+                        .expect("failed to decommit linear memory pages");
+                }
+            }
         }
 
         instance.memories.clear();
@@ -482,51 +489,81 @@ impl InstancePool {
             decommit_table_pages(base, size).expect("failed to decommit table pages");
         }
 
-        instance.tables.clear();
-        instance.dropped_elements.clear();
+        // We've now done all of the pooling-allocator-specific
+        // teardown, so we can drop the Instance and let destructors
+        // take care of any other fields (host state, globals, etc.).
+        unsafe {
+            std::ptr::drop_in_place(instance as *mut _);
+        }
+        // The instance is now uninitialized memory and cannot be
+        // touched again until we write a fresh Instance in-place with
+        // std::ptr::write in allocate() above.
 
-        // Drop all `global` values which need a destructor, such as externref
-        // values which now need their reference count dropped.
-        instance.drop_globals();
-
-        // Drop any host state
-        instance.host_state = Box::new(());
-
-        // And finally reset the module/offsets back to their original. This
-        // should put everything back in a relatively pristine state for each
-        // fresh allocation later on.
-        instance.module = self.empty_module.clone();
-        instance.offsets = VMOffsets::new(HostPtr, &self.empty_module);
-        instance.wasm_data = &[];
-        instance.vmctx_initialized = false;
-
-        self.free_list.lock().unwrap().push(index);
+        self.index_allocator.lock().unwrap().free(SlotId(index));
     }
 
     fn set_instance_memories(
+        instance_idx: usize,
         instance: &mut Instance,
-        mut memories: impl Iterator<Item = *mut u8>,
+        memories: &MemoryPool,
+        maybe_memfds: Option<&Arc<ModuleMemFds>>,
         max_pages: u64,
     ) -> Result<(), InstantiationError> {
         let module = instance.module.as_ref();
 
         debug_assert!(instance.memories.is_empty());
 
-        for plan in
-            (&module.memory_plans.values().as_slice()[module.num_imported_memories..]).iter()
+        for (memory_index, plan) in module
+            .memory_plans
+            .iter()
+            .skip(module.num_imported_memories)
         {
+            let defined_index = module
+                .defined_memory_index(memory_index)
+                .expect("should be a defined memory since we skipped imported ones");
+
             let memory = unsafe {
                 std::slice::from_raw_parts_mut(
-                    memories.next().unwrap(),
+                    memories.get_base(instance_idx, memory_index),
                     (max_pages as usize) * (WASM_PAGE_SIZE as usize),
                 )
             };
-            instance.memories.push(
-                Memory::new_static(plan, memory, commit_memory_pages, unsafe {
-                    &mut *instance.store()
-                })
-                .map_err(InstantiationError::Resource)?,
-            );
+
+            if let Some(memfds) = maybe_memfds {
+                let image = memfds.get_memory_image(defined_index);
+                let mut slot = memories.take_memfd_slot(instance_idx, memory_index);
+                let initial_size = plan.memory.minimum * WASM_PAGE_SIZE as u64;
+
+                // If instantiation fails, we can propagate the error
+                // upward and drop the slot. This will cause the Drop
+                // handler to attempt to map the range with PROT_NONE
+                // memory, to reserve the space while releasing any
+                // stale mappings. The next use of this slot will then
+                // create a new MemFdSlot that will try to map over
+                // this, returning errors as well if the mapping
+                // errors persist. The unmap-on-drop is best effort;
+                // if it fails, then we can still soundly continue
+                // using the rest of the pool and allowing the rest of
+                // the process to continue, because we never perform a
+                // mmap that would leave an open space for someone
+                // else to come in and map something.
+                slot.instantiate(initial_size as usize, image)
+                    .map_err(|e| InstantiationError::Resource(e.into()))?;
+
+                instance.memories.push(
+                    Memory::new_static(plan, memory, None, Some(slot), unsafe {
+                        &mut *instance.store()
+                    })
+                    .map_err(InstantiationError::Resource)?,
+                );
+            } else {
+                instance.memories.push(
+                    Memory::new_static(plan, memory, Some(commit_memory_pages), None, unsafe {
+                        &mut *instance.store()
+                    })
+                    .map_err(InstantiationError::Resource)?,
+                );
+            }
         }
 
         debug_assert!(instance.dropped_data.is_empty());
@@ -568,17 +605,6 @@ impl InstancePool {
     }
 }
 
-impl Drop for InstancePool {
-    fn drop(&mut self) {
-        unsafe {
-            for i in 0..self.max_instances {
-                let ptr = self.mapping.as_mut_ptr().add(i * self.instance_size) as *mut Instance;
-                std::ptr::drop_in_place(ptr);
-            }
-        }
-    }
-}
-
 /// Represents a pool of WebAssembly linear memories.
 ///
 /// A linear memory is divided into accessible pages and guard pages.
@@ -591,6 +617,10 @@ impl Drop for InstancePool {
 #[derive(Debug)]
 struct MemoryPool {
     mapping: Mmap,
+    // If using the memfd allocation scheme, the MemFd slots. We
+    // dynamically transfer ownership of a slot to a Memory when in
+    // use.
+    memfd_slots: Vec<Mutex<Option<MemFdSlot>>>,
     // The size, in bytes, of each linear memory's reservation plus the guard
     // region allocated for it.
     memory_size: usize,
@@ -675,8 +705,18 @@ impl MemoryPool {
         let mapping = Mmap::accessible_reserved(0, allocation_size)
             .context("failed to create memory pool mapping")?;
 
+        let num_memfd_slots = if cfg!(memfd) {
+            max_instances * max_memories
+        } else {
+            0
+        };
+        let memfd_slots: Vec<_> = std::iter::repeat_with(|| Mutex::new(None))
+            .take(num_memfd_slots)
+            .collect();
+
         let pool = Self {
             mapping,
+            memfd_slots,
             memory_size,
             initial_memory_offset,
             max_memories,
@@ -691,17 +731,54 @@ impl MemoryPool {
         Ok(pool)
     }
 
-    fn get(&self, instance_index: usize) -> impl Iterator<Item = *mut u8> {
+    fn get_base(&self, instance_index: usize, memory_index: MemoryIndex) -> *mut u8 {
         debug_assert!(instance_index < self.max_instances);
+        let memory_index = memory_index.as_u32() as usize;
+        debug_assert!(memory_index < self.max_memories);
+        let idx = instance_index * self.max_memories + memory_index;
+        let offset = self.initial_memory_offset + idx * self.memory_size;
+        unsafe { self.mapping.as_mut_ptr().offset(offset as isize) }
+    }
 
-        let base: *mut u8 = unsafe {
-            self.mapping.as_mut_ptr().add(
-                self.initial_memory_offset + instance_index * self.memory_size * self.max_memories,
-            ) as _
-        };
+    fn get<'a>(&'a self, instance_index: usize) -> impl Iterator<Item = *mut u8> + 'a {
+        (0..self.max_memories)
+            .map(move |i| self.get_base(instance_index, MemoryIndex::from_u32(i as u32)))
+    }
 
-        let size = self.memory_size;
-        (0..self.max_memories).map(move |i| unsafe { base.add(i * size) })
+    /// Take ownership of the given memfd slot. Must be returned via
+    /// `return_memfd_slot` when the instance is done using it.
+    fn take_memfd_slot(&self, instance_index: usize, memory_index: MemoryIndex) -> MemFdSlot {
+        let idx = instance_index * self.max_memories + (memory_index.as_u32() as usize);
+        let maybe_slot = self.memfd_slots[idx].lock().unwrap().take();
+
+        maybe_slot.unwrap_or_else(|| {
+            MemFdSlot::create(
+                self.get_base(instance_index, memory_index) as *mut c_void,
+                0,
+                self.memory_size,
+            )
+        })
+    }
+
+    /// Return ownership of the given memfd slot.
+    fn return_memfd_slot(&self, instance_index: usize, memory_index: MemoryIndex, slot: MemFdSlot) {
+        assert!(!slot.is_dirty());
+        let idx = instance_index * self.max_memories + (memory_index.as_u32() as usize);
+        *self.memfd_slots[idx].lock().unwrap() = Some(slot);
+    }
+}
+
+impl Drop for MemoryPool {
+    fn drop(&mut self) {
+        // Clear the `clear_no_drop` flag (i.e., ask to *not* clear on
+        // drop) for all MemFdSlots, and then drop them here. This is
+        // valid because the one `Mmap` that covers the whole region
+        // can just do its one munmap.
+        for mut memfd in std::mem::take(&mut self.memfd_slots) {
+            if let Some(memfd_slot) = memfd.get_mut().unwrap() {
+                memfd_slot.no_clear_on_drop();
+            }
+        }
     }
 }
 
@@ -786,7 +863,7 @@ struct StackPool {
     stack_size: usize,
     max_instances: usize,
     page_size: usize,
-    free_list: Mutex<Vec<usize>>,
+    index_allocator: Mutex<PoolingAllocationState>,
 }
 
 #[cfg(all(feature = "async", unix))]
@@ -829,25 +906,29 @@ impl StackPool {
             stack_size,
             max_instances,
             page_size,
-            free_list: Mutex::new((0..max_instances).collect()),
+            // We always use a `NextAvailable` strategy for stack
+            // allocation. We don't want or need an affinity policy
+            // here: stacks do not benefit from being allocated to the
+            // same compiled module with the same image (they always
+            // start zeroed just the same for everyone).
+            index_allocator: Mutex::new(PoolingAllocationState::new(
+                PoolingAllocationStrategy::NextAvailable,
+                max_instances,
+            )),
         })
     }
 
-    fn allocate(
-        &self,
-        strategy: PoolingAllocationStrategy,
-    ) -> Result<wasmtime_fiber::FiberStack, FiberStackError> {
+    fn allocate(&self) -> Result<wasmtime_fiber::FiberStack, FiberStackError> {
         if self.stack_size == 0 {
             return Err(FiberStackError::NotSupported);
         }
 
         let index = {
-            let mut free_list = self.free_list.lock().unwrap();
-            if free_list.is_empty() {
+            let mut alloc = self.index_allocator.lock().unwrap();
+            if alloc.is_empty() {
                 return Err(FiberStackError::Limit(self.max_instances as u32));
             }
-            let free_index = strategy.next(free_list.len());
-            free_list.swap_remove(free_index)
+            alloc.alloc(None).index()
         };
 
         debug_assert!(index < self.max_instances);
@@ -893,7 +974,7 @@ impl StackPool {
 
         decommit_stack_pages(bottom_of_stack as _, stack_size).unwrap();
 
-        self.free_list.lock().unwrap().push(index);
+        self.index_allocator.lock().unwrap().free(SlotId(index));
     }
 }
 
@@ -904,7 +985,6 @@ impl StackPool {
 /// Note: the resource pools are manually dropped so that the fault handler terminates correctly.
 #[derive(Debug)]
 pub struct PoolingInstanceAllocator {
-    strategy: PoolingAllocationStrategy,
     module_limits: ModuleLimits,
     // This is manually drop so that the pools unmap their memory before the page fault handler drops.
     instances: mem::ManuallyDrop<InstancePool>,
@@ -929,7 +1009,7 @@ impl PoolingInstanceAllocator {
             bail!("the instance count limit cannot be zero");
         }
 
-        let instances = InstancePool::new(&module_limits, &instance_limits, tunables)?;
+        let instances = InstancePool::new(strategy, &module_limits, &instance_limits, tunables)?;
 
         #[cfg(all(feature = "uffd", target_os = "linux"))]
         let _fault_handler = imp::PageFaultHandler::new(&instances)?;
@@ -937,7 +1017,6 @@ impl PoolingInstanceAllocator {
         drop(stack_size); // suppress unused warnings w/o async feature
 
         Ok(Self {
-            strategy,
             module_limits,
             instances: mem::ManuallyDrop::new(instances),
             #[cfg(all(feature = "async", unix))]
@@ -976,7 +1055,7 @@ unsafe impl InstanceAllocator for PoolingInstanceAllocator {
         &self,
         req: InstanceAllocationRequest,
     ) -> Result<InstanceHandle, InstantiationError> {
-        self.instances.allocate(self.strategy, req)
+        self.instances.allocate(req)
     }
 
     unsafe fn initialize(
@@ -990,7 +1069,7 @@ unsafe impl InstanceAllocator for PoolingInstanceAllocator {
         cfg_if::cfg_if! {
             if #[cfg(all(feature = "uffd", target_os = "linux"))] {
                 match &module.memory_initialization {
-                    wasmtime_environ::MemoryInitialization::Paged{ out_of_bounds, .. } => {
+                    wasmtime_environ::MemoryInitialization::Paged { .. } => {
                         if !is_bulk_memory {
                             super::check_init_bounds(instance, module)?;
                         }
@@ -1000,13 +1079,6 @@ unsafe impl InstanceAllocator for PoolingInstanceAllocator {
 
                         // Don't initialize the memory; the fault handler will back the pages when accessed
 
-                        // If there was an out of bounds access observed in initialization, return a trap
-                        if *out_of_bounds {
-                            return Err(InstantiationError::Trap(crate::traphandlers::Trap::wasm(
-                                wasmtime_environ::TrapCode::HeapOutOfBounds,
-                            )));
-                        }
-
                         Ok(())
                     },
                     _ => initialize_instance(instance, module, is_bulk_memory)
@@ -1023,7 +1095,7 @@ unsafe impl InstanceAllocator for PoolingInstanceAllocator {
 
     #[cfg(all(feature = "async", unix))]
     fn allocate_fiber_stack(&self) -> Result<wasmtime_fiber::FiberStack, FiberStackError> {
-        self.stacks.allocate(self.strategy)
+        self.stacks.allocate()
     }
 
     #[cfg(all(feature = "async", unix))]
@@ -1343,21 +1415,6 @@ mod test {
         );
     }
 
-    #[test]
-    fn test_next_available_allocation_strategy() {
-        let strat = PoolingAllocationStrategy::NextAvailable;
-        assert_eq!(strat.next(10), 9);
-        assert_eq!(strat.next(5), 4);
-        assert_eq!(strat.next(1), 0);
-    }
-
-    #[test]
-    fn test_random_allocation_strategy() {
-        let strat = PoolingAllocationStrategy::Random;
-        assert!(strat.next(100) < 100);
-        assert_eq!(strat.next(1), 0);
-    }
-
     #[cfg(target_pointer_width = "64")]
     #[test]
     fn test_instance_pool() -> Result<()> {
@@ -1377,6 +1434,7 @@ mod test {
         let instance_limits = InstanceLimits { count: 3 };
 
         let instances = InstancePool::new(
+            PoolingAllocationStrategy::NextAvailable,
             &module_limits,
             &instance_limits,
             &Tunables {
@@ -1390,7 +1448,10 @@ mod test {
         assert_eq!(instances.instance_size, region::page::size());
         assert_eq!(instances.max_instances, 3);
 
-        assert_eq!(&*instances.free_list.lock().unwrap(), &[0, 1, 2]);
+        assert_eq!(
+            instances.index_allocator.lock().unwrap().testing_freelist(),
+            &[SlotId(0), SlotId(1), SlotId(2)]
+        );
 
         let mut handles = Vec::new();
         let module = Arc::new(Module::default());
@@ -1399,48 +1460,49 @@ mod test {
         for _ in (0..3).rev() {
             handles.push(
                 instances
-                    .allocate(
-                        PoolingAllocationStrategy::NextAvailable,
-                        InstanceAllocationRequest {
-                            module: module.clone(),
-                            image_base: 0,
-                            functions,
-                            imports: Imports {
-                                functions: &[],
-                                tables: &[],
-                                memories: &[],
-                                globals: &[],
-                            },
-                            shared_signatures: VMSharedSignatureIndex::default().into(),
-                            host_state: Box::new(()),
-                            store: StorePtr::empty(),
-                            wasm_data: &[],
+                    .allocate(InstanceAllocationRequest {
+                        module: &module,
+                        unique_id: None,
+                        image_base: 0,
+                        functions,
+                        imports: Imports {
+                            functions: &[],
+                            tables: &[],
+                            memories: &[],
+                            globals: &[],
                         },
-                    )
+                        shared_signatures: VMSharedSignatureIndex::default().into(),
+                        host_state: Box::new(()),
+                        store: StorePtr::empty(),
+                        wasm_data: &[],
+                        memfds: None,
+                    })
                     .expect("allocation should succeed"),
             );
         }
 
-        assert_eq!(&*instances.free_list.lock().unwrap(), &[]);
+        assert_eq!(
+            instances.index_allocator.lock().unwrap().testing_freelist(),
+            &[]
+        );
 
-        match instances.allocate(
-            PoolingAllocationStrategy::NextAvailable,
-            InstanceAllocationRequest {
-                module: module.clone(),
-                functions,
-                image_base: 0,
-                imports: Imports {
-                    functions: &[],
-                    tables: &[],
-                    memories: &[],
-                    globals: &[],
-                },
-                shared_signatures: VMSharedSignatureIndex::default().into(),
-                host_state: Box::new(()),
-                store: StorePtr::empty(),
-                wasm_data: &[],
+        match instances.allocate(InstanceAllocationRequest {
+            module: &module,
+            unique_id: None,
+            functions,
+            image_base: 0,
+            imports: Imports {
+                functions: &[],
+                tables: &[],
+                memories: &[],
+                globals: &[],
             },
-        ) {
+            shared_signatures: VMSharedSignatureIndex::default().into(),
+            host_state: Box::new(()),
+            store: StorePtr::empty(),
+            wasm_data: &[],
+            memfds: None,
+        }) {
             Err(InstantiationError::Limit(3)) => {}
             _ => panic!("unexpected error"),
         };
@@ -1449,7 +1511,10 @@ mod test {
             instances.deallocate(&handle);
         }
 
-        assert_eq!(&*instances.free_list.lock().unwrap(), &[2, 1, 0]);
+        assert_eq!(
+            instances.index_allocator.lock().unwrap().testing_freelist(),
+            &[SlotId(2), SlotId(1), SlotId(0)]
+        );
 
         Ok(())
     }
@@ -1559,17 +1624,26 @@ mod test {
         assert_eq!(pool.page_size, native_page_size);
 
         assert_eq!(
-            &*pool.free_list.lock().unwrap(),
-            &[0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
+            pool.index_allocator.lock().unwrap().testing_freelist(),
+            &[
+                SlotId(0),
+                SlotId(1),
+                SlotId(2),
+                SlotId(3),
+                SlotId(4),
+                SlotId(5),
+                SlotId(6),
+                SlotId(7),
+                SlotId(8),
+                SlotId(9)
+            ],
         );
 
         let base = pool.mapping.as_ptr() as usize;
 
         let mut stacks = Vec::new();
         for i in (0..10).rev() {
-            let stack = pool
-                .allocate(PoolingAllocationStrategy::NextAvailable)
-                .expect("allocation should succeed");
+            let stack = pool.allocate().expect("allocation should succeed");
             assert_eq!(
                 ((stack.top().unwrap() as usize - base) / pool.stack_size) - 1,
                 i
@@ -1577,12 +1651,9 @@ mod test {
             stacks.push(stack);
         }
 
-        assert_eq!(&*pool.free_list.lock().unwrap(), &[]);
+        assert_eq!(pool.index_allocator.lock().unwrap().testing_freelist(), &[]);
 
-        match pool
-            .allocate(PoolingAllocationStrategy::NextAvailable)
-            .unwrap_err()
-        {
+        match pool.allocate().unwrap_err() {
             FiberStackError::Limit(10) => {}
             _ => panic!("unexpected error"),
         };
@@ -1592,8 +1663,19 @@ mod test {
         }
 
         assert_eq!(
-            &*pool.free_list.lock().unwrap(),
-            &[9, 8, 7, 6, 5, 4, 3, 2, 1, 0],
+            pool.index_allocator.lock().unwrap().testing_freelist(),
+            &[
+                SlotId(9),
+                SlotId(8),
+                SlotId(7),
+                SlotId(6),
+                SlotId(5),
+                SlotId(4),
+                SlotId(3),
+                SlotId(2),
+                SlotId(1),
+                SlotId(0)
+            ],
         );
 
         Ok(())
@@ -1666,11 +1748,6 @@ mod test {
     #[cfg(all(unix, target_pointer_width = "64", feature = "async"))]
     #[test]
     fn test_stack_zeroed() -> Result<()> {
-        // https://github.com/bytecodealliance/wasmtime/pull/2518#issuecomment-747280133
-        if std::env::var("WASMTIME_TEST_NO_HOG_MEMORY").is_ok() {
-            return Ok(());
-        }
-
         let allocator = PoolingInstanceAllocator::new(
             PoolingAllocationStrategy::NextAvailable,
             ModuleLimits {
